use crate::parser::Tokens;

macro_rules! build_ops {
    ( $name:ident, $( $(#[$docs:meta])* $op:ident),* ) => {
        #[derive(Clone, PartialEq, Eq, PartialOrd, Ord)]
        /// All of the operators in shedron
        pub enum $name {
            $(
                $(
                    #[$docs]
                )*
                $op($crate::ast::Expr, $crate::ast::Expr),
            )*
        }

        impl $name {
            /// # Build an empty operator from a token
            ///
            /// ## Behaviour
            ///
            /// `None` if `t.ty()` is not equal to [`TokenType::Op`]
            /// 
            /// [`TokenType::Op`]: $crate::parser::TokenType
            pub fn from_token(t: Tokens) -> Option<Self> {
                match t {
                    $(
                        $crate::parser::Tokens::$op => std::option::Option::Some(Ops::$op($crate::ast::Expr::Empty, $crate::ast::Expr::Empty)),
                    )*
                    _ => std::option::Option::None
                }
            }

            pub fn set_lhs(&mut self, lhs: $crate::ast::Expr) {
                match self {
                    $(
                        Ops::$op(l, _) => *l = lhs,
                    )*
                }
            }

            pub fn set_rhs(&mut self, rhs: $crate::ast::Expr) {
                match self {
                    $(
                        Ops::$op(_, r) => *r = rhs,
                    )*
                }
            }

            pub fn get_lhs(&self) -> &'_ $crate::ast::Expr {
                match self {
                    $(
                        Ops::$op(l, _) => l,
                    )*
                }
            }

            pub fn get_rhs(&self) -> &'_ $crate::ast::Expr {
                match self {
                    $(
                        Ops::$op(_, r) => r,
                    )*
                }
            }

            pub fn name(&self) -> std::string::String {
                match self {
                    $(
                        Ops::$op(_, _) => stringify!($op).to_string(),
                    )*
                }
            }

            pub fn to_expr(self) -> $crate::ast::Expr {
                Expr::Op(Box::new(self))
            }
        }
    };
}

#[derive(PartialEq, Eq, PartialOrd, Ord)]
/// An AST built from parsing shedron source code
pub struct TopLevel {
    /// Root of the tree
    pub expr: Expr,
}

impl TopLevel {
    /// Construct a new AST from a single expression
    pub fn new(expr: Expr) -> Self {
        Self { expr }
    }
}

impl std::fmt::Debug for TopLevel {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.debug_tuple("toplevel").field(&self.expr).finish()
    }
}

impl std::fmt::Display for TopLevel {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        let content = format!("{}", self.expr);

        f.write_str(content.as_str())
    }
}

build_ops! { 
    Ops,

    /// <h1>Lexical repr</h1>
    /// 
    /// `+`
    /// 
    /// <h1>Behaviour</h1>
    /// 
    /// `X:INT + Y:INT -> INT`
    /// 
    /// Adds `Y` to `X`
    /// 
    /// `X:INT[] + Y:INT -> INT[]`
    /// 
    /// Adds `Y` to the end of `X` 
    Add,

    /// <h1>Lexical repr</h1>
    /// 
    /// `-`
    /// 
    /// <h1>Behaviour</h1>
    /// 
    /// `X:INT - Y:INT -> INT`
    /// 
    /// Subtracts `Y` from `X`
    /// 
    /// `X:INT[] - Y:INT -> INT[]`
    /// 
    /// Adds `-Y` to the end of `X` 
    Sub,

    /// <h1>Lexical repr</h1>
    /// 
    /// `*`
    /// 
    /// <h1>Behaviour</h1>
    /// 
    /// `X:INT * Y:INT -> INT`
    /// 
    /// Multiplies `X` by `Y`
    /// 
    /// `X:INT[] * Y:INT -> INT`
    /// 
    /// Multiplies all `INT`s in `X` by `Y`
    Mul,

    /// <h1>Lexical repr</h1>
    /// 
    /// `/`
    /// 
    /// <h1>Behaviour</h1>
    /// 
    /// `X:INT / Y:INT -> INT`
    /// 
    /// Divides `X` by `Y`
    /// 
    /// `X:INT[] / Y:INT -> INT`
    /// 
    /// Divides all `INT`s in `X` by `Y`
    RDiv,

    /// <h1>Lexical repr</h1>
    /// 
    /// `//`
    /// 
    /// <h1>Behaviour</h1>
    /// 
    /// `X:INT // Y:INT -> INT`
    /// 
    /// Divides `X` by `Y`, rounding towards negative infinity
    /// 
    /// `X:INT[] // Y:INT -> INT`
    /// 
    /// Divides all `INT`s in `X` by `Y`, rounding towards negative infinity
    FDiv,

    /// <h1>Lexical repr</h1>
    /// 
    /// `+=`
    /// 
    /// <h1>Behaviour</h1>
    /// 
    /// `X:INT + Y:INT -> INT`
    /// 
    /// Adds `Y` to `X`
    /// 
    /// `X:INT[] + Y:INT -> INT[]`
    /// 
    /// Adds `Y` to all `INT`s in `X`
    AddEq,

    /// <h1>Lexical repr</h1>
    /// 
    /// `-=`
    /// 
    /// <h1>Behaviour</h1>
    /// 
    /// `X:INT - Y:INT -> INT`
    /// 
    /// Subtracts `Y` from `X`
    /// 
    /// `X:INT[] - Y:INT -> INT[]`
    /// 
    /// Subtracts `Y` from all `INT`s in `X`
    SubEq,

    /// <h1>Lexical repr</h1>
    /// 
    /// `^`
    ///
    /// <h1>Behaviour</h1>
    /// 
    /// `X:INT ^ Y:INT -> INT`
    /// 
    /// Returns the largest of the two passed `INT`s
    /// 
    /// `X:INT[] ^ Y:INT -> INT`
    /// 
    /// Return `Y` largest `INT`s from `X`
    Max,

    /// <h1>Lexical repr</h1>
    /// 
    /// `_`
    /// 
    /// <h1>Behaviour</h1>
    /// 
    /// `X:INT _ Y:INT -> INT`
    /// 
    /// Returns the smallest of the two passed `INT`s
    /// 
    /// `X:INT[] _ Y:INT -> INT`
    /// 
    /// Return `Y` smallest `INT`s from `X`
    Min,

    /// <h1>Lexical repr</h1>
    /// 
    /// `d`
    /// 
    /// <h1>Behaviour</h1>
    /// 
    /// `X:INT d Y:INT -> INT[]`
    /// 
    /// Rolls `X` dice in the bounds `1..=Y`
    Roll
}

#[derive(Clone, PartialEq, Eq, PartialOrd, Ord)]
/// Representation of shedron syntactical constructs
pub enum Expr {
    /// Represents an operator
    Op(Box<Ops>),

    /// Represents a sum unary operator (`[ ]`)
    Sum(Box<Expr>),

    /// Represents a drop lowest postfix operator (`-L`)
    DropLowest(Box<Expr>),

    /// Represents a drop highest postfix operator (`-H`)
    DropHighest(Box<Expr>),

    /// Represents a expression wrapped in parenthesis (`(...)`)
    Parens(Box<Expr>),

    /// Represents a number
    Number(i128),

    /// Empty // only used when initialising an empty operator, should not appear in the AST
    Empty,
}

impl Expr {
    /// Returns `true` if the expr is [`Op`].
    ///
    /// [`Op`]: Expr::Op
    #[must_use]
    pub fn is_op(&self) -> bool {
        matches!(self, Self::Op(..))
    }

    /// Returns `true` if the expr is [`Sum`].
    ///
    /// [`Sum`]: Expr::Sum
    #[must_use]
    pub fn is_sum(&self) -> bool {
        matches!(self, Self::Sum(..))
    }

    /// Returns `true` if the expr is [`Number`].
    ///
    /// [`Number`]: Expr::Number
    #[must_use]
    pub fn is_number(&self) -> bool {
        matches!(self, Self::Number(..))
    }
}

impl std::fmt::Debug for Expr {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            Expr::Op(e) => {
                let op = e.as_ref();

                match op {
                    Ops::Add(a, b) => f.debug_tuple("add").field(a).field(b).finish(),
                    Ops::Sub(a, b) => f.debug_tuple("sub").field(a).field(b).finish(),
                    Ops::Mul(a, b) => f.debug_tuple("mul").field(a).field(b).finish(),
                    Ops::FDiv(a, b) => f.debug_tuple("fdiv").field(a).field(b).finish(),
                    Ops::RDiv(a, b) => f.debug_tuple("rdiv").field(a).field(b).finish(),
                    Ops::AddEq(a, b) => f.debug_tuple("add_eq").field(a).field(b).finish(),
                    Ops::SubEq(a, b) => f.debug_tuple("sub_eq").field(a).field(b).finish(),
                    Ops::Max(a, b) => f.debug_tuple("max").field(a).field(b).finish(),
                    Ops::Min(a, b) => f.debug_tuple("min").field(a).field(b).finish(),
                    Ops::Roll(a, b) => f.debug_tuple("roll").field(a).field(b).finish(),
                }
            }
            Expr::Sum(n) => f.debug_tuple("sum").field(n).finish(),
            Expr::DropLowest(n) => f.debug_tuple("dropLowest").field(n).finish(),
            Expr::DropHighest(n) => f.debug_tuple("dropHighest").field(n).finish(),
            Expr::Parens(e) => f.debug_tuple("paren").field(e).finish(),
            Expr::Number(n) => f.debug_tuple("num").field(n).finish(),
            Expr::Empty => unreachable!(),
            
        }

        // f.write_str(content.as_str())
    }
}

impl std::fmt::Display for Expr {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        let content = match self {
            Expr::Number(n) => format!("{n}"),
            Expr::Parens(e) => format!("({e})"),
            Expr::Sum(e) => format!("[{e}]"),
            Expr::DropLowest(e) => format!("{e}-L"),
            Expr::DropHighest(e) => format!("{e}-H"),
            Expr::Op(e) => {
                let op = &**e;
                match op {
                    Ops::Add(a, b) => format!("{a}+{b}"),
                    Ops::Sub(a, b) => format!("{a}-{b}"),
                    Ops::Mul(a, b) => format!("{a}*{b}"),
                    Ops::FDiv(a, b) => format!("{a}//{b}"),
                    Ops::RDiv(a, b) => format!("{a}/{b}"),
                    Ops::AddEq(a, b) => format!("{a}+={b}"),
                    Ops::SubEq(a, b) => format!("{a}-={b}"),
                    Ops::Max(a, b) => format!("{a}^{b}"),
                    Ops::Min(a, b) => format!("{a}_{b}"),
                    Ops::Roll(a, b) => format!("{a}d{b}"),
                }
            },
            Expr::Empty => unreachable!("ooops?"),
        };

        f.write_str(content.as_str())
    }
}