use rand::{distributions::Uniform, Rng};

use crate::{
    ast::{Expr, Ops, TopLevel},
    error::RunError,
};

/// Module level result for returning `T` or a [`RunError`]
///
/// [`RunError`]: crate::error::RunError
pub type Result<T> = std::result::Result<T, RunError>;

#[derive(Debug)]
/// Final output of completely evaluated shedron
pub struct Output {
    pub rolls: Vec<i128>,
}

#[derive(Debug)]
/// All possible output values of an evaluated expression
pub enum OutputValues {
    Number(i128),
    Rolls(Vec<i128>),
}

impl OutputValues {
    /// Returns `true` if the output values is [`Rolls`].
    ///
    /// [`Rolls`]: OutputValues::Rolls
    #[must_use]
    pub fn is_rolls(&self) -> bool {
        matches!(self, Self::Rolls(..))
    }
}

/// Simple unit struct to contain interpreter function definitions
pub struct Interpreter;

impl Interpreter {
    pub fn new() -> Self {
        Self
    }

    /// Runs shedron from an AST
    pub fn run(&self, ast: TopLevel) -> Result<Output> {
        let expr = ast.expr;
        let out = self.eval_expr(expr)?;

        let rolls = match out {
            OutputValues::Number(n) => vec![n],
            OutputValues::Rolls(r) => r,
        };

        Ok(Output { rolls })
    }

    /// Evaluates a single expression to [`Result<OutputValues>`]
    ///
    /// [`OutputValues`]: Result<crate::interpret::OutputValues>
    pub fn eval_expr(&self, expr: Expr) -> Result<OutputValues> {
        match expr {
            crate::ast::Expr::Op(op) => {
                let op = op.as_ref();

                self.eval_op(op)
            }
            crate::ast::Expr::Parens(e) => self.eval_expr(*e),
            crate::ast::Expr::Sum(e) => {
                let expr = self.eval_expr(*e)?;
                match &expr {
                    OutputValues::Rolls(r) => Ok(OutputValues::Number(r.into_iter().sum())),
                    OutputValues::Number(..) => Ok(expr),
                }
            }
            crate::ast::Expr::DropLowest(e) => {
                let expr = self.eval_expr(*e)?;
                match expr {
                    OutputValues::Rolls(mut r) => {
                        r.sort_by(|a, b| a.cmp(b));
                        r.remove(0);
                        Ok(OutputValues::Rolls(r))
                    }
                    OutputValues::Number(..) => {
                        Err(RunError::new("Cannot drop elements from single number"))
                    }
                }
            }
            crate::ast::Expr::DropHighest(e) => {
                let expr = self.eval_expr(*e)?;
                match expr {
                    OutputValues::Rolls(mut r) => {
                        r.sort_by(|a, b| b.cmp(a));
                        r.remove(0);
                        Ok(OutputValues::Rolls(r))
                    }
                    OutputValues::Number(..) => {
                        Err(RunError::new("Cannot drop elements from single number"))
                    }
                }
            }
            crate::ast::Expr::Number(n) => Ok(OutputValues::Number(n)),
            crate::ast::Expr::Empty => unreachable!("oh no"),
        }
    }

    /// Evaluates a single operator to [`Result<OutputValues>`]
    ///
    /// [`OutputValues`]: Result<crate::interpret::OutputValues>
    pub fn eval_op(&self, op: &Ops) -> Result<OutputValues> {
        match op {
            Ops::Add(l, r) => {
                let l = self.eval_expr(l.clone())?;
                let r = self.eval_expr(r.clone())?;

                match (l, r) {
                    (OutputValues::Number(l), OutputValues::Number(r)) => {
                        Ok(OutputValues::Number(l + r))
                    }
                    (OutputValues::Rolls(mut l), OutputValues::Number(r)) => {
                        l.push(r);
                        Ok(OutputValues::Rolls(l))
                    }
                    (l, r) => Err(RunError::argument(op, l, r)),
                }
            }
            Ops::Sub(l, r) => {
                let l = self.eval_expr(l.clone())?;
                let r = self.eval_expr(r.clone())?;

                match (l, r) {
                    (OutputValues::Number(l), OutputValues::Number(r)) => {
                        Ok(OutputValues::Number(l - r))
                    }
                    (OutputValues::Rolls(mut l), OutputValues::Number(r)) => {
                        l.push(-r);
                        Ok(OutputValues::Rolls(l))
                    }
                    (l, r) => Err(RunError::argument(op, l, r)),
                }
            }
            Ops::Mul(l, r) => {
                let l = self.eval_expr(l.clone())?;
                let r = self.eval_expr(r.clone())?;

                match (l, r) {
                    (OutputValues::Number(l), OutputValues::Number(r)) => {
                        Ok(OutputValues::Number(l * r))
                    }
                    (OutputValues::Rolls(l), OutputValues::Number(r)) => {
                        let l = l.into_iter().map(|n| n * r).collect();
                        Ok(OutputValues::Rolls(l))
                    }
                    (l, r) => Err(RunError::argument(op, l, r)),
                }
            }
            Ops::RDiv(l, r) => {
                let l = self.eval_expr(l.clone())?;
                let r = self.eval_expr(r.clone())?;

                match (l, r) {
                    (OutputValues::Number(l), OutputValues::Number(r)) => {
                        Ok(OutputValues::Number(l / r))
                    }
                    (OutputValues::Rolls(l), OutputValues::Number(r)) => {
                        let l = l.into_iter().map(|n| n / r).collect();
                        Ok(OutputValues::Rolls(l))
                    }
                    (l, r) => Err(RunError::argument(op, l, r)),
                }
            }
            Ops::FDiv(l, r) => {
                let l = self.eval_expr(l.clone())?;
                let r = self.eval_expr(r.clone())?;

                match (l, r) {
                    (OutputValues::Number(l), OutputValues::Number(r)) => {
                        Ok(OutputValues::Number(l / r))
                    }
                    (OutputValues::Rolls(l), OutputValues::Number(r)) => {
                        let l = l.into_iter().map(|n| n / r).collect();
                        Ok(OutputValues::Rolls(l))
                    }
                    (l, r) => Err(RunError::argument(op, l, r)),
                }
            }
            Ops::AddEq(l, r) => {
                let l = self.eval_expr(l.clone())?;
                let r = self.eval_expr(r.clone())?;

                match (l, r) {
                    #[cfg(feature = "map-operators-do-single")]
                    (OutputValues::Number(l), OutputValues::Number(r)) => {
                        Ok(OutputValues::Number(l + r))
                    }
                    (OutputValues::Rolls(l), OutputValues::Number(r)) => {
                        let l = l.into_iter().map(|n| n + r).collect();
                        Ok(OutputValues::Rolls(l))
                    }
                    (l, r) => Err(RunError::argument(op, l, r)),
                }
            }
            Ops::SubEq(l, r) => {
                let l = self.eval_expr(l.clone())?;
                let r = self.eval_expr(r.clone())?;

                match (l, r) {
                    #[cfg(feature = "map-operators-do-single")]
                    (OutputValues::Number(l), OutputValues::Number(r)) => {
                        Ok(OutputValues::Number(l - r))
                    }
                    (OutputValues::Rolls(l), OutputValues::Number(r)) => {
                        let l = l.into_iter().map(|n| n - r).collect();
                        Ok(OutputValues::Rolls(l))
                    }
                    (l, r) => Err(RunError::argument(op, l, r)),
                }
            }
            Ops::Max(l, r) => {
                let l = self.eval_expr(l.clone())?;
                let r = self.eval_expr(r.clone())?;

                match (l, r) {
                    (OutputValues::Number(l), OutputValues::Number(r)) => {
                        Ok(OutputValues::Number(if l > r { l } else { r }))
                    }
                    (OutputValues::Rolls(mut l), OutputValues::Number(r)) => {
                        l.sort_by(|a, b| b.cmp(a));
                        let l = l.into_iter().take(r as usize).collect::<Vec<_>>();

                        Ok(OutputValues::Rolls(l))
                    }
                    (l, r) => Err(RunError::argument(op, l, r)),
                }
            }
            Ops::Min(l, r) => {
                let l = self.eval_expr(l.clone())?;
                let r = self.eval_expr(r.clone())?;

                match (l, r) {
                    (OutputValues::Number(l), OutputValues::Number(r)) => {
                        Ok(OutputValues::Number(if l < r { l } else { r }))
                    }
                    (OutputValues::Rolls(mut l), OutputValues::Number(r)) => {
                        l.sort_by(|a, b| a.cmp(b));
                        let l = l.into_iter().take(r as usize).collect::<Vec<_>>();

                        Ok(OutputValues::Rolls(l))
                    }
                    (l, r) => Err(RunError::argument(op, l, r)),
                }
            }
            Ops::Roll(l, r) => {
                let l = self.eval_expr(l.clone())?;
                let r = self.eval_expr(r.clone())?;

                match (l, r) {
                    (OutputValues::Number(l), OutputValues::Number(r)) => {
                        if l <= 0 {
                            return Err(RunError::new(
                                "Left hand side of roll is less than or equal to 0.",
                            ));
                        } else if r <= 0 {
                            return Err(RunError::new(
                                "Right hand side of roll is less than or equal to 0.",
                            ));
                        }
                        if r == 1 {
                            let vals: Vec<i128> = (0..l).map(|_| 1).collect();
                            Ok(OutputValues::Rolls(vals))
                        } else {
                            let mut rng = rand::thread_rng();
                            let range = Uniform::new_inclusive(1, r);

                            let vals: Vec<i128> = (0..l).map(|_| rng.sample(&range)).collect();

                            Ok(OutputValues::Rolls(vals))
                        }
                    }

                    (l, r) => Err(RunError::argument(op, l, r)),
                }
            }
        }
    }
}

#[cfg(test)]
mod test {
    use std::ops::RangeInclusive;

    use super::*;

    impl Output {
        fn expect_len(&self, len: usize) -> &Self {
            assert_eq!(
                self.rolls.len(),
                len,
                "output length is not expected {self:?}.len() != {len}",
            );
            self
        }

        fn expect_len_in(&self, len: RangeInclusive<usize>) -> &Self {
            assert!(
                len.contains(&self.rolls.len()),
                "output length is not in the expected range {self:?}.len() != {len:?}"
            );
            self
        }

        fn expect_value_at(&self, pos: usize, value_range: RangeInclusive<i128>) -> &Self {
            let v = self.rolls[pos];
            assert!(
                value_range.contains(&v),
                "value {v} at posistion {pos} was not in range {value_range:?}"
            );

            self
        }

        fn expect_values_in(&self, value_range: RangeInclusive<i128>) -> &Self {
            for (pos, _) in self.rolls.iter().enumerate() {
                self.expect_value_at(pos, value_range.clone());
            }

            self
        }

        fn expect_values(&self, value_ranges: Vec<RangeInclusive<i128>>) -> &Self {
            self.expect_len(value_ranges.len());
            for (pos, value_range) in value_ranges.iter().enumerate() {
                self.expect_value_at(pos, value_range.clone());
            }

            self
        }

        fn expect_values_at(&self, at: usize, value_ranges: Vec<RangeInclusive<i128>>) -> &Self {
            for (offset, value_range) in value_ranges.iter().enumerate() {
                let pos = at + offset;
                self.expect_value_at(pos, value_range.clone());
            }

            self
        }

        fn consume_expected_values(&self, value_ranges: Vec<RangeInclusive<i128>>) -> Self {
            dbg!(&self.rolls);
            for (pos, value_range) in value_ranges.iter().enumerate() {
                self.expect_value_at(pos, value_range.clone());
            }

            let mut rolls: Vec<i128> = self.rolls.clone();
            rolls.drain(0..value_ranges.len());
            Output { rolls: rolls }
        }
    }

    fn eval_dice_unwrap(src: &str) -> Output {
        println!("parsing {src}");
        let mut lexer = crate::parser::Lexer::new(src);
        let interpreter = Interpreter::new();
        let ast_result = lexer.parse();
        println!("evaluating {src}");
        let result = interpreter.run(ast_result.unwrap());
        println!("evaluated {src}");
        dbg!(&result);
        result.unwrap()
    }

    fn repeat_range(n: usize, range: RangeInclusive<i128>) -> Vec<RangeInclusive<i128>> {
        return std::iter::repeat(range).take(n).collect();
    }

    #[test]
    fn test_run_blind() {
        eval_dice_unwrap("2d1");
        eval_dice_unwrap("[1]");
        eval_dice_unwrap("2d[1d2]");
        eval_dice_unwrap("(10d6)+5");
        eval_dice_unwrap("[1d6]d[1d6]");
        eval_dice_unwrap("([6d([1d6])]_96*[1d6])//7");
    }

    #[test]
    fn test_expect_dice() {
        eval_dice_unwrap("[1]")
            .expect_len(1)
            .expect_values_in(1..=1);

        eval_dice_unwrap("2d6")
            .expect_len(2)
            .expect_values_in(1..=6);

        eval_dice_unwrap("([1d5]+1)d6")
            .expect_len_in(2..=6)
            .expect_values_in(1..=6);

        eval_dice_unwrap("[1d5+=1]d6")
            .expect_len_in(2..=6)
            .expect_values_in(1..=6);

        eval_dice_unwrap("([2d2]+1)d6")
            .expect_len_in(3..=6)
            .expect_values_in(1..=6);

        eval_dice_unwrap("[2d2+1]d6")
            .expect_len_in(3..=6)
            .expect_values_in(1..=6);

        eval_dice_unwrap("[2d2+=1]d6")
            .expect_len_in(3..=5)
            .expect_values_in(1..=6);

        eval_dice_unwrap("2d2+2")
            .expect_len(3)
            .consume_expected_values(repeat_range(2, 1..=2))
            .consume_expected_values(vec![2..=2])
            .expect_len(0);
        eval_dice_unwrap("2d2-2")
            .expect_len(3)
            .consume_expected_values(repeat_range(2, 1..=2))
            .consume_expected_values(vec![-2..=-2])
            .expect_len(0);

        eval_dice_unwrap("2d2+=2")
            .expect_len(2)
            .expect_values_in(1..=4);

        eval_dice_unwrap("[2d2+2]")
            .expect_len(1)
            .expect_values_in(4..=6);
        eval_dice_unwrap("[2d2]+2")
            .expect_len(1)
            .expect_values_in(4..=6);
        eval_dice_unwrap("[2d2+=2]")
            .expect_len(1)
            .expect_values_in(6..=8);
        eval_dice_unwrap("2d2*2")
            .expect_len(2)
            .expect_values_in(2..=4);
        eval_dice_unwrap("[2d2*2]")
            .expect_len(1)
            .expect_values_in(4..=8);

        eval_dice_unwrap("20d([1d3]+=3)")
            .expect_len(20)
            .expect_values_in(1..=6);

        eval_dice_unwrap("200d([1d3+2])")
            .expect_len(200)
            .expect_values_in(1..=5);

        eval_dice_unwrap("20d([1d3])+=2")
            .expect_len(20)
            .expect_values_in(3..=6);

        eval_dice_unwrap("200d4")
            .expect_len(200)
            .expect_values_in(1..=4);
        eval_dice_unwrap("200d4+=2")
            .expect_len(200)
            .expect_values_in(3..=6);
        eval_dice_unwrap("6d4-L")
            .expect_len(5)
            .expect_values_in(1..=4);
    }

    #[test]
    fn test_negative() {
        eval_dice_unwrap("1d6-1")
            .expect_values(vec![1..=6, -1..=-1]);
    }

    #[test]
    fn test_append() {
        eval_dice_unwrap("6d4+1+2+3+4")
            .consume_expected_values(repeat_range(6, 1..=4))
            .consume_expected_values(vec![1..=1])
            .consume_expected_values(vec![2..=2])
            .consume_expected_values(vec![3..=3])
            .consume_expected_values(vec![4..=4])
            .expect_len(0);
    }

    #[test]
    fn test_drop() {
        eval_dice_unwrap("15d20-L-L-L-L-L")
            .expect_len(10)
            .expect_values_in(1..=20);
        eval_dice_unwrap("15d20-D-D-D-D-D")
            .expect_len(10)
            .expect_values_in(1..=20);
    }
    #[test]
    #[should_panic]
    fn test_fail_len() {
        eval_dice_unwrap("1d6+2").expect_len(1);
    }

    #[test]
    #[should_panic]
    fn test_fail_range() {
        eval_dice_unwrap("[2]").expect_values_in(1..=1);
    }

    #[test]
    #[should_panic]
    fn test_fail_index_out_of_bounds() {
        eval_dice_unwrap("2d6").expect_value_at(5, 1..=6);
    }
}
